# Copyright contributors to the Terratorch project

"""
This is just an example of a possible structure to include SMP models
Right now it always returns a UNET, but could easily be extended to many of the models provided by SMP.
"""

import importlib

import torch
from torch import nn

from terratorch.models.model import Model, ModelFactory, ModelOutput
from terratorch.models.utils import extract_prefix_keys
from terratorch.registry import MODEL_FACTORY_REGISTRY
from terratorch.registry import BACKBONE_REGISTRY

def freeze_module(module: nn.Module):
    for param in module.parameters():
        param.requires_grad_(False)

@MODEL_FACTORY_REGISTRY.register
class GenericModelFactory(ModelFactory):

    def build_model(
        self,
        backbone: str | None = None,
        in_channels: int = 6,
        pretrained: str | bool | None = True,
        **kwargs,
    ) -> Model:
        """Factory to create models from any custom module.

        Args:
            model (str): The name for the model class.
            in_channels (int): Number of input channels.
            pretrained(str | bool): Which weights to use for the backbone. If true, will use "imagenet". If false or None, random weights. Defaults to True.

        Returns:
            Model: A wrapped generic model.
        """

        model_kwargs = _extract_prefix_keys(kwargs, "backbone_")

        try:
            model = BACKBONE_REGISTRY.build(backbone, **model_kwargs)

        except KeyError:
            raise KeyError(f"Model {backbone} not found in the registry.")

        return GenericModelWrapper(model)

class GenericModelWrapper(Model, nn.Module):
    """
    A wrapper to adapt a generic model to be used with TerraTorch

    Args: 
        model (torch.nn.Module): The model we want to wrap. 

    Returns:
        The wrapped model.
    """

    def __init__(self, model) -> None:
        super().__init__()
        self.model = model

    def freeze_encoder(self):
        freeze_module(self.model)

    def freeze_decoder(self):
        freeze_module(self.model)

    def forward(self, *args, **kwargs):
        """
        The forward method is prepared to receive any argument or keyword the wrapped model
        need to run. 
        """
        # It supposes the input has dimension (B, C, H, W)
        input_data = [args[0]] # It adapts the input to became a list of time 'snapshots'

        model_output = self.model(*args, **kwargs)

        model_output = ModelOutput(model_output)

        return model_output

def _extract_prefix_keys(d: dict, prefix: str) -> dict:
    extracted_dict = {}
    keys_to_del = []
    for k, v in d.items():
        if k.startswith(prefix):
            extracted_dict[k.split(prefix)[1]] = v
            keys_to_del.append(k)

    for k in keys_to_del:
        del d[k]

    return extracted_dict




from collections.abc import Callable, KeysView, Mapping, Set

import timm
import torch
from torch import nn

from terratorch.registry import BACKBONE_REGISTRY
from terratorch.utils import remove_unexpected_prefix

class TimmBackboneWrapper_(nn.Module):
    def __init__(self, timm_module: nn.Module) -> None:
        super().__init__()
        self._timm_module = timm_module
        # for backwards compatibility for times before necks
        self.prepare_features_for_image_model = getattr(self._timm_module, "prepare_features_for_image_model", lambda x: x)

    @property
    def out_channels(self):
        return self._timm_module.feature_info.channels()


    def forward(self, *args, **kwargs) -> list[torch.Tensor]:
        return self._timm_module(*args, **kwargs)

class TimmBackboneWrapper(nn.Module):
    def __init__(self, timm_module: nn.Module) -> None:
        super().__init__()
        self._modules.update(timm_module._modules)
        self._out_channels = timm_module.feature_info.channels()
        # for backwards compatibility for times before necks
        self.prepare_features_for_image_model = getattr(timm_module, "prepare_features_for_image_model", lambda x: x)
        self.forward = timm_module.forward
        if hasattr(timm_module, "freeze"):
            self.freeze = timm_module.freeze
    @property
    def out_channels(self):
        return self._out_channels

class TimmRegistry(Set):
    """Registry wrapper for timm"""

    def register(self, constructor: Callable | type) -> Callable:
        raise NotImplementedError()

    def build(self, name: str, features_only=True, *constructor_args, **constructor_kwargs) -> nn.Module:
        """Build and return the component.
        Use prefixes ending with _ to forward to a specific source
        """
        try:
            return TimmBackboneWrapper(
                timm.create_model(
                    name,
                    *constructor_args,
                    features_only=features_only,
                    **constructor_kwargs,
                )
            )
        except RuntimeError as e:
            if "Unknown model" in str(e):
                msg = f"Unknown model {name}"
                raise KeyError(msg) from e
            raise e

    def __iter__(self):
        return iter(timm.list_models())

    def __len__(self):
        return len(timm.list_models())

    def __contains__(self, key):
        return key in timm.list_models()

    # def __getitem__(self, name):
    #     return timm.model_entrypoint(name)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        return f"timm registry with {len(self)} registered backbones"



TIMM_BACKBONE_REGISTRY = TimmRegistry()
BACKBONE_REGISTRY.register_source("timm", TIMM_BACKBONE_REGISTRY)




import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial
from torch import Tensor
from torchmetrics import MetricCollection
from torchmetrics.classification import (
    MultilabelAccuracy,
    MultilabelFBetaScore,
)

from terratorch.models.model import ModelOutput
from terratorch.tasks import ClassificationTask


# from geobench
def _balanced_binary_cross_entropy_with_logits(outputs: Tensor, targets: Tensor) -> Tensor:
    """Compute balance binary cross entropy for multi-label classification.

    Args:
        outputs: model outputs
        targets: targets to compute binary cross entropy on
    """
    classes = targets.shape[-1]
    outputs = outputs.view(-1, classes)
    targets = targets.view(-1, classes).float()
    loss = F.binary_cross_entropy_with_logits(outputs, targets, reduction="none")
    loss = loss[targets == 0].mean() + loss[targets == 1].mean()
    return loss


class MultiLabelClassificationTask(ClassificationTask):
    def configure_losses(self) -> None:
        if self.hparams["loss"] == "bce":
            self.criterion: nn.Module = nn.BCEWithLogitsLoss()
        elif self.hparams["loss"] == "balanced_bce":
            self.criterion = _balanced_binary_cross_entropy_with_logits
        else:
            super().configure_losses()

    def configure_metrics(self) -> None:
        metrics = MetricCollection(
            {
                "Overall_Accuracy": MultilabelAccuracy(
                    num_labels=self.hparams["model_args"]["num_classes"], average="micro"
                ),
                "Average_Accuracy": MultilabelAccuracy(
                    num_labels=self.hparams["model_args"]["num_classes"], average="macro"
                ),
                "Multilabel_F1_Score": MultilabelFBetaScore(
                    num_labels=self.hparams["model_args"]["num_classes"], beta=1.0, average="micro"
                ),
            }
        )

        self.train_metrics = metrics.clone(prefix="train/")
        self.val_metrics = metrics.clone(prefix="val/")
        if self.hparams["test_dataloaders_names"] is not None:
            self.test_metrics = nn.ModuleList(
                [metrics.clone(prefix=f"test/{dl_name}/") for dl_name in self.hparams["test_dataloaders_names"]]
            )
        else:
            self.test_metrics = nn.ModuleList([metrics.clone(prefix="test/")])

    @staticmethod
    def to_multilabel_prediction(y: ModelOutput) -> Tensor:
        y_hat = y.output
        return torch.sigmoid(y_hat)

    def training_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k:batch[k] for k in other_keys}

        model_output: ModelOutput = self(x, **rest)
        loss = self.train_loss_handler.compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.train_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = self.to_multilabel_prediction(model_output)
        self.train_metrics.update(y_hat, y)

        return loss["loss"]

    def validation_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> None:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k:batch[k] for k in other_keys}
        model_output: ModelOutput = self(x, **rest)
        loss = self.val_loss_handler.compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.val_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = self.to_multilabel_prediction(model_output)
        self.val_metrics.update(y_hat, y)

    def test_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> None:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k:batch[k] for k in other_keys}
        model_output: ModelOutput = self(x, **rest)
        if dataloader_idx >= len(self.test_loss_handler):
            msg = "You are returning more than one test dataloader but not defining enough test_dataloaders_names."
            raise ValueError(msg)
        loss = self.test_loss_handler[dataloader_idx].compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.test_loss_handler[dataloader_idx].log_loss(
            partial(self.log, add_dataloader_idx=False),  # We don't need the dataloader idx as prefixes are different
            loss_dict=loss,
            batch_size=y.shape[0],
        )
        y_hat = self.to_multilabel_prediction(model_output)
        self.test_metrics[dataloader_idx].update(y_hat, y)

    def predict_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch["image"]
        file_names = batch["filename"] if "filename" in batch else None
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k: batch[k] for k in other_keys}
        model_output = self(x, **rest)
        y_hat = self.to_multilabel_prediction(model_output)
        return y_hat, file_names



"""This module contains pre-training tasks."""

import torch
import matplotlib.pyplot as plt
from torch import nn
from lightning.pytorch.callbacks import Callback
from torchgeo.datasets.utils import unbind_samples
from torchgeo.trainers import BaseTask
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MetricCollection

from terratorch.models.model import Model, ReconstructionOutput
from terratorch.registry import MODEL_FACTORY_REGISTRY
from terratorch.tasks.loss_handler import LossHandler
from terratorch.tasks.tiled_inference import TiledInferenceParameters, tiled_inference
from terratorch.tasks.regression_tasks import IgnoreIndexLossWrapper, IgnoreIndexMetricWrapper, RootLossWrapper

BATCH_IDX_FOR_VALIDATION_PLOTTING = 10


class ReconstructionTask(BaseTask):
    """
    Task with image reconstruction.
    """

    def __init__(
            self,
            model_factory: str,
            model_args: dict,
            loss: str = "mse",
            ignore_index: int | float = torch.nan,
            masked_metric: bool = True,
            lr: float = 0.001,
            optimizer: str | None = None,
            optimizer_hparams: dict | None = None,
            scheduler: str | None = None,
            scheduler_hparams: dict | None = None,
            freeze_encoder: bool = False,  # noqa: FBT001, FBT002
            freeze_decoder: bool = False,  # noqa: FBT001, FBT002
            plot_on_val: bool | int = 10,
            tiled_inference_parameters: TiledInferenceParameters | None = None,
            modalities: list[str] | None = None
    ) -> None:
        """Constructor

        Args:
            model_factory (str): Name of ModelFactory class to be used to instantiate the model.
            model_args (Dict): Arguments passed to the model factory.
            loss (str, optional): Loss to be used. Currently, supports 'mse', 'rmse', 'mae' or 'huber' loss.
                Defaults to "mse".
            aux_loss (dict[str, float] | None, optional): Auxiliary loss weights.
                Should be a dictionary where the key is the name given to the loss
                and the value is the weight to be applied to that loss.
                The name of the loss should match the key in the dictionary output by the model's forward
                method containing that output. Defaults to None.
            ignore_index (int | float, optional): Values to ignore in the metric computation. Defaults to torch.nan.
            masked_metric (bool, optional): Whether to mask the pred for the metric computation, so that metrics are
                only computed on masked pixels. Expects masks with 0 == not masked and 1 = masked. Defaults to True.
            lr (float, optional): Learning rate to be used. Defaults to 0.001.
            optimizer (str | None, optional): Name of optimizer class from torch.optim to be used.
                If None, will use Adam. Defaults to None. Overriden by config / cli specification through LightningCLI.
            optimizer_hparams (dict | None): Parameters to be passed for instantiation of the optimizer.
                Overriden by config / cli specification through LightningCLI.
            scheduler (str, optional): Name of Torch scheduler class from torch.optim.lr_scheduler
                to be used (e.g. ReduceLROnPlateau). Defaults to None.
                Overriden by config / cli specification through LightningCLI.
            scheduler_hparams (dict | None): Parameters to be passed for instantiation of the scheduler.
                Overriden by config / cli specification through LightningCLI.
            freeze_encoder (bool, optional): Whether to freeze the encoder. Defaults to False.
            freeze_decoder (bool, optional): Whether to freeze the decoder and segmentation head. Defaults to False.
            plot_on_val (bool | int, optional): Whether to plot visualizations on validation.
                If true, log every epoch. Defaults to 10. If int, will plot every plot_on_val epochs.
            tiled_inference_parameters (TiledInferenceParameters | None, optional): Inference parameters
                used to determine if inference is done on the whole image or through tiling.
            modalities list(str), optional: List of modality names that are reconstructed. Expect reconstructions as
                dict with modelity names as keys and tensors as values. Computes metrics for every modality.
                Currently, only supports image modalities (regression), no segmentation maps or similar.
        """
        self.tiled_inference_parameters = tiled_inference_parameters
        self.model_factory = MODEL_FACTORY_REGISTRY.build(model_factory)
        self.modalities = modalities
        self.ignore_index = ignore_index
        self.masked_metric = masked_metric
        super().__init__()
        self.train_loss_handler = LossHandler('train/')
        self.val_loss_handler = LossHandler('val/')
        self.test_loss_handler = LossHandler('test/')
        self.monitor = f"val/loss"
        self.plot_on_val = int(plot_on_val)

    # overwrite early stopping
    def configure_callbacks(self) -> list[Callback]:
        return []

    def configure_models(self) -> None:
        self.model: Model = self.model_factory.build_model(**self.hparams["model_args"])

        if self.hparams["freeze_encoder"]:
            if self.hparams.get("peft_config", None) is not None:
                msg = "PEFT should be run with freeze_encoder = False and freeze_decoder = False."
                raise ValueError(msg)
            self.model.freeze_encoder()
        if self.hparams["freeze_decoder"]:
            self.model.freeze_decoder()

    def configure_metrics(self) -> None:
        """Initialize the performance metrics."""

        def instantiate_metrics():
            # TODO: Handle segmentation outputs (e.g. for LULC maps with multimodal models)
            metrics = {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
            }

            return {name: IgnoreIndexMetricWrapper(metric, ignore_index=self.hparams["ignore_index"])
                    for name, metric in metrics.items()}

        if self.modalities:
            self.train_metrics = nn.ModuleDict(
                {modality: MetricCollection(instantiate_metrics(), prefix=f"train/{modality}_")
                 for modality in self.modalities}
            )
            self.val_metrics = nn.ModuleDict(
                {modality: MetricCollection(instantiate_metrics(), prefix=f"val/{modality}_")
                 for modality in self.modalities}
            )
            self.test_metrics = nn.ModuleDict(
                {modality: MetricCollection(instantiate_metrics(), prefix=f"test/{modality}_")
                 for modality in self.modalities}
            )
        else:
            self.train_metrics = MetricCollection(instantiate_metrics(), prefix="train/")
            self.val_metrics = MetricCollection(instantiate_metrics(), prefix="val/")
            self.test_metrics = MetricCollection(instantiate_metrics(), prefix="test/")

    def mask_input(self, x: torch.Tensor | dict, mask: torch.Tensor | dict) -> torch.Tensor:
        """
        Args:
            x: input tensor or dict of input tensors, [B, C, H, W] or [B, C, T, H, W].
            mask: mask tensor or dict of mask tensors, [B, H, W] or [B, T, H, W].
        """
        if isinstance(x, dict):
            # Multimodal data
            for key in x.keys():
                x[key] = self.mask_input(x[key], mask[key])
        else:
            x = x.clone()
            if len(x.shape) == 4:
                # Adding channel dim to mask
                mask = mask.unsqueeze(1).repeat(1, x.shape[1], 1, 1)
            elif len(x.shape) == 5:
                # Temporal data
                mask = mask.unsqueeze(1).repeat(1, x.shape[1], 1, 1, 1)
            if x.dtype == torch.uint8:
                # Convert for NaN ignore_index
                x = x.to(torch.float32)
            # Mask input
            x[mask == True] = self.ignore_index

        return x

    def training_step(self, batch: dict, batch_idx: int, dataloader_idx: int = 0) -> torch.Tensor:
        """Compute the train loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        if isinstance(x, dict):
            batch_size = list(x.values())[0].shape[0]
        elif isinstance(x, torch.Tensor):
            batch_size = x.shape[0]
        else:
            raise ValueError("Could not infer batch size from input data.")

        other_keys = batch.keys() - {"image", "mask", "filename"}
        rest = {k: batch[k] for k in other_keys}

        output: ReconstructionOutput | tuple = self(x, **rest)
        if not isinstance(output, ReconstructionOutput):
            output = ReconstructionOutput(*output)

        self.train_loss_handler.log_loss(self.log, loss_dict=output.loss, batch_size=batch_size)

        if self.masked_metric and output.mask is not None:
            x = self.mask_input(x, output.mask)

        if isinstance(x, dict):
            # Multimodal data
            for modality in x.keys():
                if modality in self.train_metrics:
                    self.train_metrics[modality].update(output.pred[modality], x[modality])
        else:
            # Single modality
            self.train_metrics.update(output.pred, x)

        return output.loss["loss"]

    def on_train_epoch_end(self) -> None:
        if isinstance(self.test_metrics, MetricCollection):
            self.log_dict(self.train_metrics.compute(), sync_dist=True)
            self.train_metrics.reset()
        else:
            for metric in self.train_metrics.values():
                self.log_dict(metric.compute(), sync_dist=True)
                metric.reset()
        return super().on_train_epoch_end()

    def _do_plot_samples(self, batch_index):
        if not self.plot_on_val:  # dont plot if self.plot_on_val is 0
            return False

        return (
            batch_index < BATCH_IDX_FOR_VALIDATION_PLOTTING
            and hasattr(self.trainer, "datamodule")
            and self.logger
            and not self.current_epoch % self.plot_on_val  # will be True every self.plot_on_val epochs
            and hasattr(self.logger, "experiment")
            and (hasattr(self.logger.experiment, "add_figure") or hasattr(self.logger.experiment, "log_figure"))
        )

    def validation_step(self, batch: dict, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the validation loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        if isinstance(x, dict):
            batch_size = list(x.values())[0].shape[0]
        elif isinstance(x, torch.Tensor):
            batch_size = x.shape[0]
        else:
            raise ValueError("Could not infer batch size from input data.")

        other_keys = batch.keys() - {"image", "mask", "filename"}
        rest = {k: batch[k] for k in other_keys}

        output: ReconstructionOutput | tuple = self(x, **rest)
        if not isinstance(output, ReconstructionOutput):
            output = ReconstructionOutput(*output)

        self.val_loss_handler.log_loss(self.log, loss_dict=output.loss, batch_size=batch_size)

        if self.masked_metric and output.mask is not None:
            x = self.mask_input(x, output.mask)

        if isinstance(x, dict):
            if not isinstance(self.val_metrics, nn.ModuleDict):
                raise ValueError(f'Multimodal data provided but no image modalities. '
                                 f'Please provide modalities (list of str) to the ReconstructionTask.')
            # Multimodal data
            for modality in x.keys():
                if modality in self.val_metrics:
                    self.val_metrics[modality].update(output.pred[modality], x[modality])
        else:
            # Single modality
            self.val_metrics.update(output.pred, x)

        if self._do_plot_samples(batch_idx):
            try:
                datamodule = self.trainer.datamodule
                if isinstance(batch["image"], dict):
                    # Multimodal input
                    rgb_modality = getattr(datamodule, 'rgb_modality', None) or list(batch["image"].keys())[0]
                    batch["image"] = batch["image"][rgb_modality]
                    batch["prediction"] = output.pred[rgb_modality].cpu()
                    if output.mask is not None:
                        batch["mask"] = output.mask[rgb_modality].cpu()
                else:
                    batch["prediction"] = output.pred.cpu()
                    batch["mask"] = output.mask.cpu()
                batch['image'] = batch['image'].cpu()
                sample = unbind_samples(batch)[0]
                fig = datamodule.val_dataset.plot(sample)
                if fig:
                    summary_writer = self.logger.experiment
                    if hasattr(summary_writer, "add_figure"):
                        summary_writer.add_figure(f"image/{batch_idx}", fig, global_step=self.global_step)
                    elif hasattr(summary_writer, "log_figure"):
                        summary_writer.log_figure(
                            self.logger.run_id, fig, f"epoch_{self.current_epoch}_{batch_idx}.png"
                        )
            except ValueError:
                pass
            finally:
                plt.close()

    def on_validation_epoch_end(self) -> None:
        if isinstance(self.test_metrics, MetricCollection):
            self.log_dict(self.val_metrics.compute(), sync_dist=True)
            self.val_metrics.reset()
        else:
            for metric in self.val_metrics.values():
                self.log_dict(metric.compute(), sync_dist=True)
                metric.reset()
        return super().on_validation_epoch_end()

    def test_step(self, batch: dict, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the test loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        if isinstance(x, dict):
            batch_size = list(x.values())[0].shape[0]
        elif isinstance(x, torch.Tensor):
            batch_size = x.shape[0]
        else:
            raise ValueError("Could not infer batch size from input data.")

        other_keys = batch.keys() - {"image", "mask", "filename"}
        rest = {k: batch[k] for k in other_keys}

        output: ReconstructionOutput | tuple = self(x, **rest)
        if not isinstance(output, ReconstructionOutput):
            output = ReconstructionOutput(*output)

        self.test_loss_handler.log_loss(self.log, loss_dict=output.loss, batch_size=batch_size)

        if self.masked_metric and output.mask is not None:
            x = self.mask_input(x, output.mask)

        if isinstance(x, dict):
            # Multimodal data
            for modality in x.keys():
                if modality in self.test_metrics:
                    self.test_metrics[modality].update(output.pred[modality], x[modality])
        else:
            # Single modality
            self.test_metrics.update(output.pred, x)

    def on_test_epoch_end(self) -> None:
        if isinstance(self.test_metrics, MetricCollection):
            self.log_dict(self.test_metrics.compute(), sync_dist=True)
            self.test_metrics.reset()
        else:
            # multiple metric dicts
            for metric in self.test_metrics.values():
                self.log_dict(metric.compute(), sync_dist=True)
                metric.reset()
        return super().on_test_epoch_end()

    def predict_step(self, batch: dict, batch_idx: int, dataloader_idx: int = 0) -> torch.Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch["image"]
        image_size = x.shape[-2:]
        file_names = batch["filename"]
        other_keys = batch.keys() - {"image", "mask", "filename"}
        rest = {k: batch[k] for k in other_keys}

        def model_forward(x):
            out = self(x)[1]
            if isinstance(out, ReconstructionOutput):
                pred  = out.pred
            else:
                pred = out[1]
            return pred

        # Tiled inference for autoencoder ? Is it making sense ?
        if self.tiled_inference_parameters:
            # TODO: tiled inference does not work with additional input data (**rest)
            pred: torch.Tensor = tiled_inference(model_forward, x, 1, self.tiled_inference_parameters)
        else:
            pred: torch.Tensor = model_forward(x)
        # TODO: Return mask and file_names?
        return pred, file_names



# Copyright (c) EPFL VILAB.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# Based on timm, DeiT, DINO, MoCo-v3, BEiT, MAE-priv and MAE code bases
# https://github.com/rwightman/pytorch-image-models/tree/master/timm
# https://github.com/facebookresearch/deit
# https://github.com/facebookresearch/dino
# https://github.com/facebookresearch/moco-v3
# https://github.com/microsoft/unilm/tree/master/beit
# https://github.com/BUPT-PRIV/MAE-priv
# https://github.com/facebookresearch/mae
# --------------------------------------------------------

import math
import warnings

import torch
import torch.nn as nn
from einops import rearrange


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


def build_2d_sincos_posemb(h, w, embed_dim=1024, temperature=10000.0):
    """Sine-cosine positional embeddings from MoCo-v3

    Source: https://github.com/facebookresearch/moco-v3/blob/main/vits.py
    """
    grid_w = torch.arange(w, dtype=torch.float32)
    grid_h = torch.arange(h, dtype=torch.float32)
    grid_w, grid_h = torch.meshgrid(grid_w, grid_h)
    assert (
        embed_dim % 4 == 0
    ), "Embed dimension must be divisible by 4 for 2D sin-cos position embedding"
    pos_dim = embed_dim // 4
    omega = torch.arange(pos_dim, dtype=torch.float32) / pos_dim
    omega = 1.0 / (temperature**omega)
    out_w = torch.einsum("m,d->md", [grid_w.flatten(), omega])
    out_h = torch.einsum("m,d->md", [grid_h.flatten(), omega])
    pos_emb = torch.cat(
        [torch.sin(out_w), torch.cos(out_w), torch.sin(out_h), torch.cos(out_h)], dim=1
    )[None, :, :]
    pos_emb = rearrange(pos_emb, "b (h w) d -> b d h w", h=h, w=w, d=embed_dim)
    return pos_emb


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2,
        )

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    # type: (Tensor, float, float, float, float) -> Tensor
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)

    def extra_repr(self) -> str:
        return "p={}".format(self.drop_prob)


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        # x = self.drop(x)
        # commit this for the orignal BERT implement
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = qkv.unbind(0)  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class CrossAttention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0.0, proj_drop=0.0):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, context):
        B, N, C = x.shape
        _, M, _ = context.shape

        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )
        kv = (
            self.kv(context)
            .reshape(B, M, 2, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):

    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class DecoderBlock(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.self_attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.cross_attn = CrossAttention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.query_norm = norm_layer(dim)
        self.context_norm = norm_layer(dim)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, context):
        x = x + self.drop_path(self.self_attn(self.norm1(x)))
        x = x + self.drop_path(
            self.cross_attn(self.query_norm(x), self.context_norm(context))
        )
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

import torch
import logging
from terratorch.registry import TERRATORCH_FULL_MODEL_REGISTRY
from huggingface_hub import hf_hub_download

logger = logging.getLogger('terramind')

try:
    from .vqvae import VQVAE, DiVAE
    vqvae_available = True
    import_error = None
except Exception as e:
    logger.debug(f'Could not import TerraMind tokenizer due to ImportError({e})')
    vqvae_available = False
    import_error = e



# Model definitions
__all__ = [
    'terramind_v1_tokenizer_s2l2a',
    'terramind_v1_tokenizer_s1rtc',
    'terramind_v1_tokenizer_s1grd',
    'terramind_v1_tokenizer_dem',
    'terramind_v1_tokenizer_lulc',
    'terramind_v1_tokenizer_ndvi',
]

pretrained_weights = {
    "terramind_v1_tokenizer_s2l2a": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S2L2A",
        "hf_hub_filename": "TerraMind_Tokenizer_S2L2A.pt",
    },
    "terramind_v1_tokenizer_s1rtc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1RTC",
        "hf_hub_filename": "TerraMind_Tokenizer_S1RTC.pt",
    },
    "terramind_v1_tokenizer_s1grd": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1GRD",
        "hf_hub_filename": "TerraMind_Tokenizer_S1GRD.pt",
    },
    "terramind_v1_tokenizer_dem": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-DEM",
        "hf_hub_filename": "TerraMind_Tokenizer_DEM.pt",
    },
    "terramind_v1_tokenizer_lulc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-LULC",
        "hf_hub_filename": "TerraMind_Tokenizer_LULC.pt",
    },
    "terramind_v1_tokenizer_ndvi": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-NDVI",
        "hf_hub_filename": "TerraMind_Tokenizer_NDVI.pt",
    },
}

def build_vqvae(
        model_type: str = 'divae',
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        **kwargs):

    if not vqvae_available:
        warnings.warn(f'Cannot import VQVAE/DiVAE from terramind. '
                      f'\nMake sure to install `pip install diffusers==0.30.0`.')
        raise import_error

    if model_type == 'divae':
        model = DiVAE(**kwargs)
    elif model_type == 'vqvae':
        model = VQVAE(**kwargs)
    else:
        raise ValueError(f'Unknown model type: {model_type}')

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]['hf_hub_id'],
                                          filename=pretrained_weights[variant]['hf_hub_filename'])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        model.load_state_dict(state_dict, strict=True)

    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s2l2a(**kwargs):
    """
    S2L2A Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type='vit_b_enc', decoder_type='unet_patched', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type='fsq', codebook_size='8-8-8-6-5', latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type='sample', beta_schedule='linear', zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn='mse', conditioning='concat', resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=1, batch_size_eval=None, epochs=100, stop_after_epoch=250, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='fp16', model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='sen2l2a@264', mask_value=None, data_path=['./data/MajorTOM-Core/train', './data/SSL4EOS12/train_single'], eval_data_path='./data/MajorTOM-Core/val', imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule='DDIMScheduler', num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=15, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir='output/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264/2k_continue', device='cuda', seed=0, resume='output/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264/2k_continue/checkpoint_29.pth', auto_resume=True, start_epoch=30, num_workers=5, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project=None, wandb_entity='erik-scheurer', wandb_run_name='2048_S2TerraMesh_divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264', wandb_tags=[], show_user_warnings=False, world_size=32, local_rank=-1, dist_on_itp=False, dist_url='env://', run_name='tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264', patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path='/p/project/geofm4eo/scheurer2/4m4eo/cfgs/default/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=32, all_domains=['sen2l2a@264'], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    tokenizer = build_vqvae(
        variant='terramind_v1_tokenizer_s2l2a',
        image_size=256,
        n_channels=12,
        encoder_type='vit_b_enc',
        decoder_type='unet_patched',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='8-8-8-6-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s1rtc(**kwargs):
    """
    S1RTC Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type='vit_b_enc', decoder_type='unet_patched', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type='fsq', codebook_size='8-8-8-6-5', latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type='sample', beta_schedule='linear', zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn='mse', conditioning='concat', resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=1, batch_size_eval=None, epochs=100, stop_after_epoch=250, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='fp16', model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='sen1@264', mask_value=None, data_path='./data/MajorTOM-Core/train/', eval_data_path='./data/MajorTOM-Core/val', imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule='DDIMScheduler', num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=1000, eval_metrics_freq=0, eval_image_log_freq=1000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir='output/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264/2k_MajorTOM224256', device='cuda', seed=0, resume='output/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264/2k_MajorTOM224256/checkpoint_14.pth', auto_resume=True, start_epoch=15, num_workers=5, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project=None, wandb_entity='erik-scheurer', wandb_run_name='2048_2k_MajorTOM224-256', wandb_tags=[], show_user_warnings=False, world_size=32, local_rank=-1, dist_on_itp=False, dist_url='env://', run_name='tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264', patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path='/p/project/geofm4eo/scheurer2/4m4eo/cfgs/default/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=32, all_domains=['sen1@264'], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    
    tokenizer = build_vqvae(
        variant='terramind_v1_tokenizer_s1rtc',
        image_size=256,
        n_channels=2,
        encoder_type='vit_b_enc',
        decoder_type='unet_patched',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='8-8-8-6-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer

@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s1grd(**kwargs):
    """
    S1GRD Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type='vit_b_enc', decoder_type='unet_patched', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type='fsq', codebook_size='8-8-8-6-5', latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type='sample', beta_schedule='linear', zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn='mse', conditioning='concat', resolution_cond=False, eval_res_cond=None, batch_size=1, batch_size_eval=None, epochs=100, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.004266666667, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='fp16', model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='sen1@264', mask_value=None, data_path='../fm-geospatial/data/SSL4EOS12/train/', eval_data_path='../fm-geospatial/data/SSL4EOS12/train/', imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, use_wds=False, s3_endpoint='', s3_data_endpoint=None, wds_n_repeats=1, wds_shuffle_buffer_tar=1000, wds_shuffle_buffer_repeat=1000, s3_multipart_chunksize_mb=512, s3_multipart_threshold_mb=512, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule='DDIMScheduler', num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=5000, eval_metrics_freq=5000, eval_image_log_freq=5000, num_logged_images=15, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir='../fm-geospatial/4m4eo-checkpoints/S1-SSL4EO-full-run/', device='cuda', seed=0, resume='', auto_resume=False, start_epoch=0, num_workers=10, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project='4m4eo', wandb_entity=None, wandb_run_name='divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264', wandb_tags=[], show_user_warnings=False, world_size=6, local_rank=-1, dist_on_itp=False, dist_url='env://', s3_path='', s3_save_dir='', run_name='tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264', patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path='cfgs/default/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=6, all_domains=['sen1@264'], input_size=256, lr=0.00010000000000781251)
    """
    
    tokenizer = build_vqvae(
        variant='terramind_v1_tokenizer_s1grd',
        image_size=256,
        n_channels=2,
        encoder_type='vit_b_enc',
        decoder_type='unet_patched',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='8-8-8-6-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_dem(**kwargs):
    """
    DEM Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=256, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type='vit_b_enc', decoder_type='unet_patched', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type='fsq', codebook_size='8-8-8-6-5', latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type='sample', beta_schedule='linear', zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn='mse', conditioning='concat', resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, stop_after_epoch=250, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='fp16', model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='dem@264', mask_value=None, data_path='./data/TerraMesh/train', eval_data_path='./data/TerraMesh/val', imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=False, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule='DDIMScheduler', num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir='artifacts/runs/sm-dem-20-epochs', device='cuda', seed=0, resume='artifacts/runs/sm-dem-20-epochs/checkpoint_12.pth', auto_resume=True, start_epoch=13, num_workers=8, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project='terramind-dem', wandb_entity='fasteo-ibm-jsc', wandb_run_name='2048_epochs_20/divae/rgb/DEM-ViTB-UNetP4_16k_224-264', wandb_run_id=None, wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url='env://', run_name='tokenization/divae/rgb/DEM-ViTB-UNetP4_16k_224-264', patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path='/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/divae/rgb/DEM-ViTB-UNetP4_16k_224-264.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=16, all_domains=['dem@264'], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    
    tokenizer = build_vqvae(
        variant='terramind_v1_tokenizer_dem',
        image_size=256,
        n_channels=1,
        encoder_type='vit_b_enc',
        decoder_type='unet_patched',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='8-8-8-6-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_lulc(**kwargs):
    """
    LULC Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=16, input_size_enc=None, input_size_dec=None, encoder_type='vit_b_enc', decoder_type='vit_b_dec', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, loss_fn='cross_entropy', percept_loss_type=None, percept_loss_blocks='blocks.2-blocks.5-blocks.8-blocks.11', percept_loss_distance='cosine', percept_loss_weight=0.0, quantizer_type='fsq', codebook_size='7-5-5-5-5', num_codebooks=1, latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0016, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='bf16', dtype_percept=None, model_ema=True, model_ema_decay=0.99, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='lulc@264', mask_value=None, data_path='./data/TerraMesh/train', eval_data_path='./data/TerraMesh/val', min_crop_scale=0.8, cache_datasets=False, num_locations=64, dist_eval=True, step_eval=True, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=1000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, keep_batches_ratio=1.0, output_dir='/p/home/jusers/maurogiovanni1/juwels/4m4eo/artifacts/runs/sm-lulc-20-00016', device='cuda', seed=0, resume='/p/home/jusers/maurogiovanni1/juwels/4m4eo/artifacts/runs/sm-lulc-20-00016/checkpoint_13.pth', auto_resume=True, start_epoch=14, num_workers=8, pin_mem=True, find_unused_params=False, log_wandb=True, wandb_project='terramind-lulc', wandb_entity='fasteo-ibm-jsc', wandb_run_name='2048_epochs_20/lr_00016/acc/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256', wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url='env://', run_name='tokenization/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256', stop_after_epoch=250, epoch_eval=False, imagenet_default_mean_and_std=False, sample_sentinel_data=True, num_seasons=1, config_path='/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=16, all_domains=['lulc@264'], input_size=256, lr=3.535533905932738e-05, effective_batch_size=2048)
    """

    tokenizer = build_vqvae(
        model_type='vqvae',
        variant='terramind_v1_tokenizer_lulc',
        image_size=256,
        n_channels=10,
        encoder_type='vit_b_enc',
        decoder_type='vit_b_dec',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='7-5-5-5-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer



@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_ndvi(**kwargs):
    """
    NDVI Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=256, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type='vit_b_enc', decoder_type='unet_patched', post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type='fsq', codebook_size='8-8-8-6-5', latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy='batch_random', commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type='sample', beta_schedule='linear', zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn='mse', conditioning='concat', resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, stop_after_epoch=250, save_ckpt_freq=1, opt='adamw', opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype='fp16', model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain='ndvi@264', mask_value=None, data_path='./data/TerraMesh/train', eval_data_path='./data/TerraMesh/val', imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=False, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule='DDIMScheduler', num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir='artifacts/runs/sm-ndvi-20-epochs', device='cuda', seed=0, resume='artifacts/runs/sm-ndvi-20-epochs/checkpoint_12.pth', auto_resume=True, start_epoch=13, num_workers=8, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project='terramind-tokenizers', wandb_entity='fasteo-ibm-jsc', wandb_run_name='2048_epochs_20/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264', wandb_run_id=None, wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url='env://', run_name='tokenization/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264', patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path='/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264.yaml', rank=0, gpu=0, distributed=True, dist_backend='nccl', num_tasks=16, all_domains=['ndvi@264'], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """

    tokenizer = build_vqvae(
        variant='terramind_v1_tokenizer_ndvi',
        image_size=256,
        n_channels=1,
        encoder_type='vit_b_enc',
        decoder_type='unet_patched',
        prediction_type='sample',
        post_mlp=True,
        patch_size=16,
        quant_type='fsq',
        codebook_size='8-8-8-6-5',
        latent_dim=5,
        **kwargs
    )

    return tokenizer



import random
import warnings
from pathlib import Path

import albumentations as A  # noqa: N812
import h5py
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F  # noqa: N812
from torchgeo.datasets import NonGeoDataset

from terratorch.datasets.utils import pad_numpy, to_tensor

CAT_TILES = ["31TBF", "31TCF", "31TCG", "31TDF", "31TDG"]
FR_TILES = ["31TCJ", "31TDK", "31TCL", "31TDM", "31UCP", "31UDR"]

MAX_TEMPORAL_IMAGE_SIZE = (366, 366)

SELECTED_CLASSES = [
    110,  # 'Wheat'
    120,  # 'Maize'
    140,  # 'Sorghum'
    150,  # 'Barley'
    160,  # 'Rye'
    170,  # 'Oats'
    330,  # 'Grapes'
    435,  # 'Rapeseed'
    438,  # 'Sunflower'
    510,  # 'Potatoes'
    770,  # 'Peas'
]


class Sen4AgriNet(NonGeoDataset):
    def __init__(
        self,
        data_root: str,
        bands: list[str] | None = None,
        scenario: str = "random",
        split: str = "train",
        transform: A.Compose = None,
        truncate_image: int | None = 4,
        pad_image: int | None = 4,
        spatial_interpolate_and_stack_temporally: bool = True,  # noqa: FBT001, FBT002
        seed: int = 42,
    ):
        """
        Pytorch Dataset class to load samples from the [Sen4AgriNet](https://github.com/Orion-AI-Lab/S4A) dataset, supporting
        multiple scenarios for splitting the data.

        Args:
            data_root (str): Root directory of the dataset.
            bands (list of str, optional): List of band names to load. Defaults to all available bands.
            scenario (str): Defines the splitting scenario to use. Options are:
                - 'random': Random split of the data.
                - 'spatial': Split by geographical regions (Catalonia and France).
                - 'spatio-temporal': Split by region and year (France 2019 and Catalonia 2020).
            split (str): Specifies the dataset split. Options are 'train', 'val', or 'test'.
            transform (albumentations.Compose, optional): Albumentations transformations to apply to the data.
            truncate_image (int, optional): Number of timesteps to truncate the time dimension of the image.
                If None, no truncation is applied. Default is 4.
            pad_image (int, optional): Number of timesteps to pad the time dimension of the image.
                If None, no padding is applied. Default is 4.
            spatial_interpolate_and_stack_temporally (bool): Whether to interpolate bands and concatenate them over time
            seed (int): Random seed used for data splitting.
        """
        self.data_root = Path(data_root) / "data"
        self.transform = transform if transform else lambda **batch: to_tensor(batch)
        self.scenario = scenario
        self.seed = seed
        self.truncate_image = truncate_image
        self.pad_image = pad_image
        self.spatial_interpolate_and_stack_temporally = spatial_interpolate_and_stack_temporally

        if bands is None:
            bands = ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B09", "B10", "B11", "B12", "B8A"]
        self.bands = bands

        self.image_files = list(self.data_root.glob("**/*.nc"))

        self.train_files, self.val_files, self.test_files = self.split_data()

        if split == "train":
            self.image_files = self.train_files
        elif split == "val":
            self.image_files = self.val_files
        elif split == "test":
            self.image_files = self.test_files

    def __len__(self):
        return len(self.image_files)

    def split_data(self):
        random.seed(self.seed)

        if self.scenario == "random":
            random.shuffle(self.image_files)
            total_files = len(self.image_files)
            train_split = int(0.6 * total_files)
            val_split = int(0.8 * total_files)

            train_files = self.image_files[:train_split]
            val_files = self.image_files[train_split:val_split]
            test_files = self.image_files[val_split:]

        elif self.scenario == "spatial":
            catalonia_files = [f for f in self.image_files if any(tile in f.stem for tile in CAT_TILES)]
            france_files = [f for f in self.image_files if any(tile in f.stem for tile in FR_TILES)]

            val_split_cat = int(0.2 * len(catalonia_files))
            train_files = catalonia_files[val_split_cat:]
            val_files = catalonia_files[:val_split_cat]
            test_files = france_files

        elif self.scenario == "spatio-temporal":
            france_files = [f for f in self.image_files if any(tile in f.stem for tile in FR_TILES)]
            catalonia_files = [f for f in self.image_files if any(tile in f.stem for tile in CAT_TILES)]

            france_2019_files = [f for f in france_files if "2019" in f.stem]
            catalonia_2020_files = [f for f in catalonia_files if "2020" in f.stem]

            val_split_france_2019 = int(0.2 * len(france_2019_files))
            train_files = france_2019_files[val_split_france_2019:]
            val_files = france_2019_files[:val_split_france_2019]
            test_files = catalonia_2020_files

        return train_files, val_files, test_files

    def __getitem__(self, index: int):
        patch_file = self.image_files[index]

        with h5py.File(patch_file, "r") as patch_data:
            output = {}
            images_over_time = []
            for band in self.bands:
                band_group = patch_data[band]
                band_data = band_group[f"{band}"][:]
                time_vector = band_group["time"][:]

                sorted_indices = np.argsort(time_vector)
                band_data = band_data[sorted_indices].astype(np.float32)

                if self.truncate_image:
                    band_data = band_data[-self.truncate_image :]
                if self.pad_image:
                    band_data = pad_numpy(band_data, self.pad_image)

                if self.spatial_interpolate_and_stack_temporally:
                    band_data = torch.from_numpy(band_data)
                    band_data = band_data.clone().detach()

                    interpolated = F.interpolate(
                        band_data.unsqueeze(0), size=MAX_TEMPORAL_IMAGE_SIZE, mode="bilinear", align_corners=False
                    ).squeeze(0)
                    images_over_time.append(interpolated)
                else:
                    output[band] = band_data

            if self.spatial_interpolate_and_stack_temporally:
                images = torch.stack(images_over_time, dim=0).numpy()
                output["image"] = images

            labels = patch_data["labels"]["labels"][:].astype(int)
            parcels = patch_data["parcels"]["parcels"][:].astype(int)

        output["mask"] = labels

        image_shape = output["image"].shape[-2:]
        mask_shape = output["mask"].shape

        if image_shape != mask_shape:
            diff_h = mask_shape[0] - image_shape[0]
            diff_w = mask_shape[1] - image_shape[1]

            output["image"] = np.pad(
                output["image"],
                [(0, 0), (0, 0), (diff_h // 2, diff_h - diff_h // 2), (diff_w // 2, diff_w - diff_w // 2)],
                mode="constant",
                constant_values=0,
            )

        linear_encoder = {val: i + 1 for i, val in enumerate(sorted(SELECTED_CLASSES))}
        linear_encoder[0] = 0

        output["image"] = output["image"].transpose(0, 2, 3, 1)
        output["mask"] = self.map_mask_to_discrete_classes(output["mask"], linear_encoder)

        if self.transform:
            output = self.transform(**output)

        output["parcels"] = parcels

        return output

    def plot(self, sample, suptitle=None, show_axes=False):
        rgb_bands = ["B04", "B03", "B02"]

        if not all(band in sample for band in rgb_bands):
            warnings.warn("No RGB image.")  # noqa: B028
            return None

        rgb_images = []
        for t in range(sample["B04"].shape[0]):
            rgb_image = torch.stack([sample[band][t] for band in rgb_bands])

            # Normalization
            rgb_min = rgb_image.min(dim=1, keepdim=True).values.min(dim=2, keepdim=True).values
            rgb_max = rgb_image.max(dim=1, keepdim=True).values.max(dim=2, keepdim=True).values
            denom = rgb_max - rgb_min
            denom[denom == 0] = 1
            rgb_image = (rgb_image - rgb_min) / denom

            rgb_image = rgb_image.permute(1, 2, 0).numpy()
            rgb_images.append(np.clip(rgb_image, 0, 1))

        dates = torch.arange(sample["B04"].shape[0])

        return self._plot_sample(rgb_images, dates, sample.get("labels"), suptitle=suptitle, show_axes=show_axes)

    def _plot_sample(self, images, dates, labels=None, suptitle=None, show_axes=False):
        num_images = len(images)
        cols = 5
        rows = (num_images + cols - 1) // cols

        fig, ax = plt.subplots(rows, cols, figsize=(20, 4 * rows))
        axes_visibility = "on" if show_axes else "off"

        for i, image in enumerate(images):
            ax[i // cols, i % cols].imshow(image)
            ax[i // cols, i % cols].set_title(f"T{i+1} - Day {dates[i].item()}")
            ax[i // cols, i % cols].axis(axes_visibility)

        if labels is not None:
            if rows * cols > num_images:
                target_ax = ax[(num_images) // cols, (num_images) % cols]
            else:
                fig.add_subplot(rows + 1, 1, 1)
                target_ax = fig.gca()

            target_ax.imshow(labels.numpy(), cmap="tab20")
            target_ax.set_title("Labels")
            target_ax.axis(axes_visibility)

        for k in range(num_images, rows * cols):
            ax[k // cols, k % cols].axis(axes_visibility)

        if suptitle:
            plt.suptitle(suptitle)

        plt.tight_layout()
        plt.show()

    def map_mask_to_discrete_classes(self, mask, encoder):
        map_func = np.vectorize(lambda x: encoder.get(x, 0))
        return map_func(mask)



import warnings
import torch
import torch.nn.functional as F  # noqa: N812
from torch import Tensor, nn

"""
Adapted from https://github.com/yassouali/pytorch-segmentation/blob/master/models/upernet.py
"""
class ConvModule(nn.Module):
    def __init__(self, in_channels, out_channels,
                 kernel_size=3, padding=0,
                 dilation=1, stride=1,
                 inplace=False, transpose=False,
                 scale_factor=None) -> None:

        super().__init__()

        if transpose:
            kind = "Transpose"
        else:
            kind = ""

        conv_name = f"Conv{kind}2d"

        if transpose:

            stride = scale_factor
            padding = (kernel_size - scale_factor) // 2

        conv_template = getattr(nn, conv_name)
        self.conv= conv_template(in_channels, out_channels, kernel_size, padding=padding, dilation=dilation, stride=stride, bias=False)
        self.norm = nn.BatchNorm2d(out_channels)
        self.act = nn.ReLU(inplace=inplace)

    def forward(self, x):
  
        return self.act(self.norm(self.conv(x)))


"""
It cames from https://github.com/open-mmlab/mmsegmentation
"""
def resize(input,
           size=None,
           scale_factor=None,
           mode='nearest',
           align_corners=None,
           warning=True):
    if warning:
        if size is not None and align_corners:
            input_h, input_w = tuple(int(x) for x in input.shape[2:])
            output_h, output_w = tuple(int(x) for x in size)
            if output_h > input_h or output_w > input_w:
                if ((output_h > 1 and output_w > 1 and input_h > 1
                     and input_w > 1) and (output_h - 1) % (input_h - 1)
                        and (output_w - 1) % (input_w - 1)):
                    warnings.warn(
                        f'When align_corners={align_corners}, '
                        'the output would more aligned if '
                        f'input size {(input_h, input_w)} is `x+1` and '
                        f'out size {(output_h, output_w)} is `nx+1`')

    return F.interpolate(input, size, scale_factor, mode, align_corners)






# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import warnings

import numpy as np
import torch
from torch import nn
from functools import partial
from einops import rearrange
import torch.nn.functional as F

from .encoder_embeddings import ImageEncoderEmbedding, ImageTokenEncoderEmbedding
from .decoder_embeddings import ImageTokenDecoderEmbedding
from .tm_utils import LayerNorm
from .modality_info import MODALITY_INFO
from .generate import GenerationSampler, build_chained_generation_schedules
from .terramind import TerraMind
from terratorch.models.backbones.terramind.tokenizer.tokenizer_register import (
    terramind_v1_tokenizer_s2l2a,
    terramind_v1_tokenizer_s1rtc,
    terramind_v1_tokenizer_s1grd,
    terramind_v1_tokenizer_dem,
    terramind_v1_tokenizer_lulc,
    terramind_v1_tokenizer_ndvi
)

def build_modality_embeddings(modalities, img_size=None, dim=None, patch_size=None):
    mod_embeddings = {}
    mod_name_mapping = {}
    for modality in modalities:
        # New modalities can be provided as {'name': <num_channels>}
        if isinstance(modality, dict):
            for key, value in modality.items():
                if isinstance(value, nn.Module):
                    mod_embeddings[key] = value
                elif isinstance(value, int):
                    mod_embeddings[key] = ImageEncoderEmbedding(num_channels=value, dim_tokens=dim, image_size=img_size,
                                                                patch_size=patch_size, sincos_pos_emb=True)
                else:
                    raise ValueError(f'Modalities must be provided as a list of strings or dicts, or as a dict with '
                                     f'the values being nn.Module or int (number of channels of the modality). '
                                     f'Found {key}: {value} ({type(value)})')
                mod_name_mapping[key] = key
            continue

        # Cover multiple naming conventions
        modality_renamed = (modality.lower()
                            .replace('s2', 'sen2')
                            .replace('s1', 'sen1')
                            .replace('text', 'caption')
                            .replace('coordinates', 'coords')
                            )

        # Get modality key in MODALITY_INFO
        if modality_renamed in MODALITY_INFO.keys():
            key = modality_renamed
        elif 'sen2l2a' in modality_renamed:
            key = 'untok_sen2l2a@224'
        elif 'sen2l1c' in modality_renamed:
            key = 'untok_sen2l1c@224'
        elif 'sen1rtc' in modality_renamed:
            key = 'untok_sen1rtc@224'
        elif 'sen1grd' in modality_renamed:
            key = 'untok_sen1grd@224'
        elif 'rgb' in modality_renamed:
            key = 'untok_sen2rgb@224'
        elif 'dem' in modality_renamed:
            key = 'untok_dem@224'
        elif 'lulc' in modality_renamed:
            key = 'tok_lulc@224'
        elif 'ndvi' in modality_renamed:
            key = 'tok_ndvi@224'
        elif 'caption' in modality_renamed:
            raise NotImplementedError('Captions are not yet supported.')
        elif 'coords' in modality_renamed:
            raise NotImplementedError('Captions are not yet supported.')
        else:
            key = modality

        if key in MODALITY_INFO.keys():
            mod_info = MODALITY_INFO[key]
            mod_embeddings[key] = mod_info['encoder_embedding'](image_size=img_size, dim_tokens=dim, **mod_info)
            mod_name_mapping[modality] = key  # Requires manual mapping for loading model weights
        else:
            raise NotImplementedError(f'Could not find modality {modality} in default modality info.')

    return mod_embeddings, mod_name_mapping


def build_output_modality_embeddings(modalities, img_size=None, dim=None, patch_size=None):
    mod_embeddings = {}
    mod_name_mapping = {}
    for modality in modalities:
        # Cover multiple naming conventions
        modality_renamed = (modality.lower()
                            .replace('s2', 'sen2')
                            .replace('s1', 'sen1')
                            .replace('text', 'caption')
                            .replace('coordinates', 'coords')
                            )

        # Get modality key in MODALITY_INFO
        if 'sen2' in modality_renamed:
            key = 'tok_sen2l2a@224'
        elif 'sen1rtc' in modality_renamed:
            key = 'tok_sen1rtc@224'
        elif 'sen1' in modality_renamed:  # Default to S1GRD if not specified
            key = 'tok_sen1grd@224'
        elif 'dem' in modality_renamed:
            key = 'tok_dem@224'
        elif 'lulc' in modality_renamed:
            key = 'tok_lulc@224'
        elif 'ndvi' in modality_renamed:
            key = 'tok_ndvi@224'
        elif 'caption' in modality_renamed:
            raise NotImplementedError('Captions are not yet supported.')
        elif 'coords' in modality_renamed:
            raise NotImplementedError('Captions are not yet supported.')
        else:
            key = modality

        if key in MODALITY_INFO.keys():
            mod_info = MODALITY_INFO[key]
            mod_embeddings[key] = mod_info['decoder_embedding'](image_size=img_size, dim_tokens=dim, **mod_info)
            mod_name_mapping[modality] = key  # Requires manual mapping for loading model weights
        else:
            raise NotImplementedError(f'Could not find modality {modality} in default modality info.'
                                      f'Available modalities: S2L2A, S1RTC, S1GRD, DEM, LULC, NDVI.')

    return mod_embeddings, mod_name_mapping



def build_tokenizer(input_modalities, output_modalities, pretrained):
    tokenizer_dict = {
        'tok_sen2l2a@224': terramind_v1_tokenizer_s2l2a,
        'tok_sen1rtc@224': terramind_v1_tokenizer_s1rtc,
        'tok_sen1grd@224': terramind_v1_tokenizer_s1grd,
        'tok_dem@224': terramind_v1_tokenizer_dem,
        'tok_lulc@224': terramind_v1_tokenizer_lulc,
        'tok_ndvi@224': terramind_v1_tokenizer_ndvi
    }
    # TODO: Add loading only encoder/decoder
    tokenizer = {}
    for modality in input_modalities:
        if modality in tokenizer_dict:
            tokenizer[modality] = tokenizer_dict[modality](pretrained=pretrained)

    for modality in output_modalities:
        if modality in tokenizer_dict and modality not in tokenizer:
            tokenizer[modality] = tokenizer_dict[modality](pretrained=pretrained)
        else:
            warnings.warn(f'Tokenizer for output modality {modality} not found.')

    tokenizer = nn.ModuleDict(tokenizer)

    return tokenizer


class TerraMindGeneration(nn.Module):
    """Modified TerraMind model for a "Thinking in Modalities" approach.

    Args:
        img_size (int): Input image size.
        modalities (list, dict, optional): List of modality keys and dicts, or dict with modality keys and values being
            ints (num_channels of modality) or nn.Module (patch embedding layer).
        output_modalities (list, optional): List of tokenized modalities used for the TiM approach. The TiM outputs are
            generated in the same order as specified in the given list. Defaults to [tbd].
        decoding_steps (list, int): Number of decoding steps for each TiM modality. Defaults to 1.
        temps (list, float): Sampling temperatures for each TiM modality. Defaults to 1.0.
        top_p (float): Top-p sampling threshold for TiM modalities. Ignored if set to 0.0. Defaults to 0.8.
        top_k (int): Top-k sampling threshold for TiM modalities. Ignored if set to 0. Defaults to 0.
        merge_method (str, optional): Specify how the output is merged for further processing. One of 'mean', 'max',
            'concat', 'dict', or None. 'mean', 'max', and 'concat' are dropping all sequence modality tokens, split all
            image modality tokens and reduce the by applying the appropriate method. 'dict' splits all tokens into a
            dictionary {'modality': torch.Tensor}. Defaults to 'mean'.
        patch_size (int): Patch size.
        in_chans (int): Number of input image channels.
        dim (int): Patch embedding dimension.
        encoder_depth (int): Depth of ViT / number of encoder blocks.
        num_heads (int): Number of attention heads in each ViT block.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool): If True, add a learnable bias to query, key, value.
        proj_bias (bool): If True, adds a bias to the attention out proj layer.
        mlp_bias (bool): If True, adds a learnable bias for the feedforward.
        drop_path_rate (float): Stochastic depth rate.
        drop_rate (float): Dropout rate.
        attn_drop_rate (float): Attention dropout rate.
        modality_drop_rate (float): Drop modality inputs during training.
        act_layer (nn.Module): Activation layer.
        norm_layer (nn.Module): Normalization layer.
        gated_mlp (bool): If True, makes the feedforward gated (e.g., for SwiGLU)
        qk_norm (bool): If True, normalizes the query and keys (as in ViT-22B)
        use_act_checkpoint (bool): If True, use activation checkpointing.
        encoder_norm (bool): If True, adds a norm layer after the last encoder block.
    """

    def __init__(
            self,
            img_size: int = 224,
            modalities: list[str] | dict[str, int | nn.Module] | None = None,
            output_modalities: list[str] | None = None,
            pretrained: bool = False,
            decoding_steps: list[int] | int = 1,
            temps: list[float] | float = 1.0,
            top_p: float = 0.8,
            top_k: int = 0,
            timesteps: int = 50,
            patch_size: int = 16,
            dim: int = 768,
            encoder_depth: int = 12,
            decoder_depth: int = 12,
            num_heads: int = 12,
            mlp_ratio: float = 4.0,
            qkv_bias: bool = True,
            proj_bias: bool = True,
            mlp_bias: bool = True,
            act_layer: torch.Tensor = nn.GELU,
            norm_layer: partial | nn.Module = partial(LayerNorm, eps=1e-6),
            gated_mlp: bool = False,
            qk_norm: bool = False,
            standardize: bool = False,
            offset: dict[str, float] | None = None,
            pretraining_mean: dict[str, list] | None = None,
            pretraining_std: dict[str, list] | None = None,
    ):
        super().__init__()

        if modalities is None or len(modalities) == 0:
            raise ValueError('Input modalities not provided.')
        elif isinstance(modalities, dict):
            modalities = [modalities]
        elif not isinstance(modalities, list):
            raise ValueError(f'Modalities must be a list of modality keys or a dict with embedding layers.')

        self.output_modalities = output_modalities or ["sen2l2a@224"]
        if isinstance(decoding_steps, list):
            assert len(decoding_steps) == len(self.output_modalities), \
                "Number of decoding steps must match number of output modalities."
        else:
            decoding_steps = [decoding_steps] * len(self.output_modalities)
        if isinstance(temps, list):
            assert len(temps) == len(self.output_modalities), \
                "Number of temperatures must match number of output modalities."
        else:
            temps = [temps] * len(self.output_modalities)

        # Parameters for generation schedule
        self.top_p = top_p
        self.top_k = top_k
        self.timesteps = timesteps
        self.decoding_steps = decoding_steps
        self.temps = temps

        # Init embeddings
        self.encoder_embeddings, mod_name_mapping = build_modality_embeddings(
            modalities, img_size=img_size, dim=dim, patch_size=patch_size)
        self.decoder_embeddings, decoder_name_mapping = build_output_modality_embeddings(
            self.output_modalities, img_size=img_size, dim=dim, patch_size=patch_size)
        self.output_modalities = list(decoder_name_mapping.values())  # Update output modality names
        self.mod_name_mapping = decoder_name_mapping | mod_name_mapping  # Merging dicts
        self.modalities = list(mod_name_mapping.keys())  # Further code expects list
        self.image_modalities = [key for key, value in self.encoder_embeddings.items()
             if isinstance(value, ImageEncoderEmbedding) or isinstance(value, ImageTokenEncoderEmbedding)]
        self.output_image_modalities = [key for key, value in self.decoder_embeddings.items()
                                        if isinstance(value, ImageTokenDecoderEmbedding)]
        self.output_mod_name_mapping = {v: k for k, v in decoder_name_mapping.items()}
        self.standardize = standardize
        self.offset = offset or {}

        if offset is not None:
            for mod, o in offset.items():
                # Add offset to mean values
                assert mod in self.mod_name_mapping, f"offset {mod} not defined in input or output modalities"
                pretraining_mean[self.mod_name_mapping[mod]] = \
                    np.array(pretraining_mean[self.mod_name_mapping[mod]], dtype=np.float32) + o

        self.pretraining_mean = {mod: torch.tensor(mean)[None, :, None, None]
                                 for mod, mean in pretraining_mean.items()} if pretraining_mean else {}
        self.pretraining_std = {mod: torch.tensor(std)[None, :, None, None]
                                for mod, std in pretraining_std.items()} if pretraining_std else {}

        # Build MAE model
        mae_model = TerraMind(
            encoder_embeddings=self.encoder_embeddings,
            decoder_embeddings=self.decoder_embeddings,
            modality_info=MODALITY_INFO,
            dim=dim,
            encoder_depth=encoder_depth,
            decoder_depth=decoder_depth,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            proj_bias=proj_bias,
            mlp_bias=mlp_bias,
            act_layer=act_layer,
            norm_layer=norm_layer,
            gated_mlp=gated_mlp,
            qk_norm=qk_norm,
        )

        self.sampler = GenerationSampler(mae_model)

        self.tokenizer = build_tokenizer(input_modalities=list(self.encoder_embeddings.keys()),
                                         output_modalities=list(self.decoder_embeddings.keys()),
                                         pretrained=pretrained)

    def to(self, device):
        super().to(device)
        # Move standardization values to device
        for mod, mean in self.pretraining_mean.items():
            self.pretraining_mean[mod] = mean.to(device)
        for mod, std in self.pretraining_std.items():
            self.pretraining_std[mod] = std.to(device)
        return self

    def forward(
            self,
            d: dict[str, torch.Tensor] | torch.Tensor | None = None,
            standardize: bool | None = None,
            offset: dict[str, float] | None = None,
            timesteps: int = None,
            verbose: bool = False,
            **kwargs
    ) -> dict[str, torch.Tensor]:
        """
        Forward pass of the model.

        Args:
            d (dict, torch.Tensor): Dict of inputs or input tensor with shape (B, C, H, W)

            Alternatively, keyword arguments with modality=tensor.

        Returns:
            dict[str, torch.Tensor]: Dict of generated images
        """
        # Handle single image modality
        if isinstance(d, torch.Tensor):
            # Assuming first modality
            d = {self.modalities[0]: d}
        elif d is None:
            d = {}
            assert len(kwargs), "No input provided."

        # Add additional keyword args to input dict
        for key, value in kwargs.items():
            d[key] = value

        # Get batch size and device
        B = d[list(d.keys())[0]].shape[0]
        device = d[list(d.keys())[0]].device

        standardize = standardize if standardize is not None else self.standardize
        if standardize:
            for mod, value in d.items():
                d[mod] = ((value - self.pretraining_mean[self.mod_name_mapping[mod]]) /
                          self.pretraining_std[self.mod_name_mapping[mod]])

        # Define the initial input
        input_dict = {}
        # Default values if no images are provided
        img_num_tokens, image_size = 196, (224, 224)
        for mod, value in d.items():
            patch_size = self.encoder_embeddings[self.mod_name_mapping[mod]].patch_size
            num_tokens = int((value.shape[-1] / patch_size[-1]) * (value.shape[-2] / patch_size[-2]))

            if self.mod_name_mapping[mod] in self.image_modalities:
                img_num_tokens = num_tokens
                image_size = (value.shape[-2], value.shape[-1])

            # Run tokenizer encoder for tokenized input modalities
            if self.mod_name_mapping[mod] in self.tokenizer:
                if 'lulc' in self.mod_name_mapping[mod]:
                    # TODO Hack: One hot encoding for LULC classes. Generalize code.
                    if len(value.shape) == 3:
                        value = F.one_hot(value.to(int), num_classes=10).permute(0, 3, 1, 2).to(torch.float32)
                    elif len(value.shape) == 4 and value.shape[1] == 1:
                        value = F.one_hot(value.to(int).squeeze(1), num_classes=10).permute(0, 3, 1, 2).to(torch.float32)
                    elif len(value.shape) == 4 and value.shape[1] == 10:
                        # Correct shape
                        pass
                    else:
                        raise ValueError('Expect LULC data with 10 classes. '
                            'Either with class indexes and shape [B, H, W] or one hot encoded [B, 10, H, W].')

                # Tokenize
                value = self.tokenizer[self.mod_name_mapping[mod]].encode(value)[2]

            input_dict[self.mod_name_mapping[mod]] = {
                "tensor": value,
                "input_mask": torch.zeros(B, num_tokens, dtype=torch.bool, device=device),
                "target_mask": torch.ones(B, num_tokens, dtype=torch.bool, device=device),
            }

        # Initialize output modalities
        sum_tokens = 0
        for mod in self.output_modalities:
            mod_num_tokens = img_num_tokens if mod in self.output_image_modalities else 196  # TODO: Not tested for sequence data.
            sum_tokens += mod_num_tokens

            if mod in input_dict:
                warnings.warn(f'The modality {mod} is used as input and output which is not possible for the same '
                              f'patches. Random sampling 50% of tokens as input and output.')
                input_dict[mod]['input_mask'] = torch.rand((B, mod_num_tokens), device=device) < 0.5
                input_dict[mod]['target_mask'] = ~input_dict[mod]['input_mask']
                input_dict[mod]['tensor'] = input_dict[mod]['tensor'].reshape(B, -1).to(torch.long)
            else:
                input_dict[mod] = {
                    'tensor': torch.zeros((B, mod_num_tokens), dtype=torch.int64, device=device),
                    'input_mask': torch.ones((B, mod_num_tokens), dtype=torch.bool, device=device),
                    'target_mask': torch.zeros((B, mod_num_tokens), dtype=torch.bool, device=device),
                }

        # Predict tokens of output modalities
        schedule = build_chained_generation_schedules(
            cond_domains=[self.mod_name_mapping[m] for m in self.modalities],
            target_domains=self.output_modalities,
            tokens_per_target=[img_num_tokens] * len(self.output_modalities),
            autoregression_schemes=["roar"] * len(self.output_modalities),
            decoding_steps=self.decoding_steps,
            token_decoding_schedules=["linear"] * len(self.output_modalities),
            temps=self.temps,
            temp_schedules=["constant"] * len(self.output_modalities),
            cfg_scales=[1.0] * len(self.output_modalities),
            cfg_schedules=["constant"] * len(self.output_modalities),
            cfg_grow_conditioning=True,
        )

        out_dict = self.sampler.generate(
            input_dict,
            schedule,
            verbose=False,
            seed=random.randint(-(2 ** 31), 2 ** 31 - 1),
            top_p=self.top_p,
            top_k=self.top_k,
            num_tokens=sum_tokens,
        )

        # TODO Vary timesteps based on codebook diversity
        timesteps = timesteps or self.timesteps
        out = {}
        for mod in self.output_modalities:
            tok = out_dict[mod]['tensor']
            if mod in self.output_image_modalities:
                patch_size = self.tokenizer[mod].patch_size
                tok = rearrange(tok, "b (nh nw) -> b nh nw",
                                nh=image_size[0] // patch_size, nw=image_size[1] // patch_size)
            out[self.output_mod_name_mapping[mod]] = self.tokenizer[mod].decode_tokens(
                tok,
                image_size=image_size,
                timesteps=timesteps,
                verbose=verbose
            )

        if standardize:
            for mod, value in out.items():
                out[mod] = (value * self.pretraining_std[self.mod_name_mapping[mod]] +
                            self.pretraining_mean[self.mod_name_mapping[mod]])

        return out



# Copyright contributors to the Terratorch project

import warnings
from torch import nn

from terratorch.models.model import ModelFactory
from terratorch.models.peft_utils import get_peft_backbone
from terratorch.registry import FULL_MODEL_REGISTRY, MODEL_FACTORY_REGISTRY


def _get_model(model: str | nn.Module, **model_kwargs) -> nn.Module:
    if isinstance(model, nn.Module):
        return model
    return FULL_MODEL_REGISTRY.build(model, **model_kwargs)


def _check_all_args_used(kwargs):
    if kwargs:
        msg = f"arguments {kwargs} were passed but not used."
        raise ValueError(msg)


@MODEL_FACTORY_REGISTRY.register
class FullModelFactory(ModelFactory):
    def build_model(
        self,
        model: str | nn.Module,
        rescale: bool = True,  # noqa: FBT002, FBT001
        padding: str = "reflect",
        peft_config: dict | None = None,
        **kwargs,
    ) -> nn.Module:
        """Generic model factory that wraps any model.

        All kwargs are passed to the model.

        Args:
            task (str): Task to be performed. Currently supports "segmentation" and "regression".
            model (str, nn.Module): Model to be used. If a string, will look for such models in the different
                registries supported (internal terratorch registry, ...). If a torch nn.Module, will use it
                directly.
            rescale (bool): Whether to apply bilinear interpolation to rescale the model output if its size
                is different from the ground truth. Only applicable to pixel wise models
                (e.g. segmentation, pixel wise regression, reconstruction). Defaults to True.
            padding (str): Padding method used if images are not divisible by the patch size. Defaults to "reflect".
            peft_config (dict): Configuration options for using [PEFT](https://huggingface.co/docs/peft/index).
                The dictionary should have the following keys:
                - "method": Which PEFT method to use. Should be one implemented in PEFT, a list is available [here](https://huggingface.co/docs/peft/package_reference/peft_types#peft.PeftType).
                - "replace_qkv": String containing a substring of the name of the submodules to replace with QKVSep.
                  This should be used when the qkv matrices are merged together in a single linear layer and the PEFT
                  method should be applied separately to query, key and value matrices (e.g. if LoRA is only desired in
                  Q and V matrices). e.g. If using Prithvi this should be "qkv"
                - "peft_config_kwargs": Dictionary containing keyword arguments which will be passed to [PeftConfig](https://huggingface.co/docs/peft/package_reference/config#peft.PeftConfig)


        Returns:
            nn.Module: Full model.
        """

        model = _get_model(model, **kwargs)

        # If patch size is not provided in the config or by the model, it might lead to errors due to irregular images.
        patch_size = kwargs.get("patch_size", None)

        if patch_size is None:
            # Infer patch size from model by checking all backbone modules
            for module in model.modules():
                if hasattr(module, "patch_size"):
                    patch_size = module.patch_size
                    break

        if peft_config is not None:
            if not kwargs.get("pretrained", False):
                msg = (
                    "You are using PEFT without a pretrained backbone. If you are loading a checkpoint afterwards "
                    "this is probably fine, but if you are training a model check the backbone_pretrained parameter."
                )
                warnings.warn(msg, stacklevel=1)

            model = get_peft_backbone(peft_config, model)

        return model



import torch 
from torch import nn
import numpy as np

from terratorch.registry import TERRATORCH_DECODER_REGISTRY
from .utils import ConvModule, resize

@TERRATORCH_DECODER_REGISTRY.register
class ASPPModule(nn.Module):
    """Atrous Spatial Pyramid Pooling (ASPP) Module.

    Args:
        dilations (tuple[int]): Dilation rate of each layer.
        in_channels (int): Input channels.
        channels (int): Channels after modules, before conv_seg.
        conv_cfg (dict|None): Config of conv layers.
        norm_cfg (dict|None): Config of norm layers.
        act_cfg (dict): Config of activation layers.
    """

    def __init__(self, dilations, in_channels, channels, conv_cfg, norm_cfg,
                 act_cfg):

        super(ASPPModule, self).__init__()

        self.dilations = dilations
        self.in_channels = in_channels
        self.channels = channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        modules_list = list()

        for dilation in dilations:

            layer_module = ConvModule(
                    self.in_channels,
                    self.channels,
                    1 if dilation == 1 else 3,
                    dilation=dilation,
                    padding=0 if dilation == 1 else dilation,)

                    # TODO Extend it to support more possible configurations
                    # for convolution, normalization and activation.
                    #conv_cfg=self.conv_cfg,
                    #norm_cfg=self.norm_cfg,
                    #act_cfg=self.act_cfg)

            modules_list.append(layer_module)

        self.model_sequence = torch.nn.ModuleList(modules_list)

    def forward(self, x):
        """Forward function."""
        outs = []
        for module in self.model_sequence:
            outs.append(module(x))

        return outs

@TERRATORCH_DECODER_REGISTRY.register
class ASPPHead(nn.Module):
    """Rethinking Atrous Convolution for Semantic Image Segmentation.

    This head is the implementation of `DeepLabV3
    <https://arxiv.org/abs/1706.05587>`_.

    Args:
        dilations (tuple[int]): Dilation rates for ASPP module.
            Default: (1, 6, 12, 18).
    """

    def __init__(self, dilations:list | tuple =(1, 6, 12, 18), 
                 in_channels:int=None, 
                 channels:int=None,
                 out_dim:int=3,
                 align_corners=False,
                 head_dropout_ratio:float=0.3,
                 input_transform: str = None,
                 in_index: int = -1,
                 **kwargs):

        super(ASPPHead, self).__init__(**kwargs)

        self.dilations = dilations
        self.in_channels = in_channels
        self.channels = channels
        self.out_dim = out_dim

        self.align_corners = align_corners
        self.input_transform = input_transform
        self.in_index = in_index 

        if 'conv_cfg' not in kwargs:
            self.conv_cfg = self._default_conv_cfg

        if 'norm_cfg' not in kwargs:
            self.norm_cfg = self._default_norm_cfg

        if 'act_cfg' not in kwargs:
            self.act_cfg = self._default_act_cfg

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels,
                self.channels,
                1,

                # TODO Extend it to support more possible configurations
                # for convolution, normalization and activation.
                #self.conv_cfg,
                #norm_cfg=self.norm_cfg,
                #act_cfg=self.act_cfg))
                ))

        self.aspp_modules = ASPPModule(
            dilations,
            self.in_channels,
            self.channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            padding=1,)
            # TODO Extend it to support more possible configurations
            # for convolution, normalization and activation.
            #conv_cfg=self.conv_cfg,
            #norm_cfg=self.norm_cfg,
            #act_cfg=self.act_cfg)

        if head_dropout_ratio > 0:
            self.dropout = nn.Dropout2d(head_dropout_ratio)

    @property
    def _default_conv_cfg(self):
        return {"kernel_size": 3, "padding": 0, "bias": False}

    @property
    def _default_norm_cfg(self):
        return {}

    @property
    def _default_act_cfg(self):
        return {}

    def _transform_inputs(self, inputs):
        """Transform inputs for decoder.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            Tensor: The transformed inputs
        """

        if self.input_transform == 'resize_concat':
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(
                    input=x,
                    size=inputs[0].shape[2:],
                    mode='bilinear',
                    align_corners=self.align_corners) for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == 'multiple_select':
            inputs = [inputs[i] for i in self.in_index]
        else:
            inputs = inputs[self.in_index]

        return inputs

    def _forward_feature(self, inputs):
        """Forward function.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            feats (Tensor): A tensor of shape (batch_size, self.channels,
                H, W) which is feature map for last layer of decoder head.
        """
        inputs = self._transform_inputs(inputs)

        aspp_outs = [
            resize(
                self.image_pool(inputs),
                size=inputs.size()[2:],
                mode='bilinear',
                align_corners=self.align_corners)
        ]

        aspp_outs.extend(self.aspp_modules(inputs))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        feats = self.bottleneck(aspp_outs)

        return feats

    def forward(self, inputs):

        output = self._forward_feature(inputs)

        return output

@TERRATORCH_DECODER_REGISTRY.register
class ASPPSegmentationHead(ASPPHead):
    """Rethinking Atrous Convolution for Semantic Image Segmentation.

    This head is the implementation of `DeepLabV3
    <https://arxiv.org/abs/1706.05587>`_.

    Args:
        dilations (tuple[int]): Dilation rates for ASPP module.
            Default: (1, 6, 12, 18).
    """

    def __init__(self, channel_list,
                 dilations:list | tuple =(1, 6, 12, 18), 
                 in_channels:int=None, 
                 channels:int=None,
                 num_classes:int=2,
                 align_corners=False,
                 head_dropout_ratio:float=0.3,
                 input_transform: str = None,
                 in_index: int = -1,
                 **kwargs):

        super(ASPPSegmentationHead, self).__init__(
                 dilations=dilations, 
                 in_channels=in_channels, 
                 channels=channels,
                 align_corners=align_corners,
                 head_dropout_ratio=head_dropout_ratio,
                 input_transform=input_transform,
                 in_index=in_index,
                **kwargs)

        self.num_classes = num_classes
        self.conv_seg = nn.Conv2d(self.channels, self.num_classes, kernel_size=1)

        if head_dropout_ratio > 0:
            self.dropout = nn.Dropout2d(head_dropout_ratio)

    def segmentation_head(self, features):

        """PixelWise classification"""

        if self.dropout is not None:
            features = self.dropout(features)
        output = self.conv_seg(features)
        return output

    def forward(self, inputs):

        output = self._forward_feature(inputs)
        output = self.segmentation_head(output)

        return output

@TERRATORCH_DECODER_REGISTRY.register
class ASPPRegressionHead(ASPPHead):
    """Rethinking Atrous Convolution for regression.

    This head is the implementation of `DeepLabV3
    <https://arxiv.org/abs/1706.05587>`_.

    Args:
        dilations (tuple[int]): Dilation rates for ASPP module.
            Default: (1, 6, 12, 18).
    """

    def __init__(self, channel_list,
                 dilations:list | tuple =(1, 6, 12, 18), 
                 in_channels:int=None, 
                 channels:int=None,
                 out_channels:int=1,
                 align_corners=False,
                 head_dropout_ratio:float=0.3,
                 input_transform: str = None,
                 in_index: int = -1,
                 **kwargs):

        super(ASPPRegressionHead, self).__init__(
                 dilations=dilations, 
                 in_channels=in_channels, 
                 channels=channels,
                 out_dim=out_channels,
                 align_corners=align_corners,
                 head_dropout_ratio=head_dropout_ratio,
                 input_transform=input_transform,
                 in_index=in_index,
                **kwargs)

        self.out_channels = out_channels
        self.conv_reg = nn.Conv2d(self.channels, self.out_channels, kernel_size=1)

    def regression_head(self, features):

        """PixelWise regression"""
        if self.dropout is not None:
            features = self.dropout(features)
        output = self.conv_reg(features)
        return output


    def forward(self, inputs):

        output = self._forward_feature(inputs)
        output = self.regression_head(output)

        return output





# Copyright contributors to the Terratorch project

import torch
import torch.nn as nn
from timm.models.vision_transformer import Block, PatchEmbed

from terratorch.registry import TERRATORCH_DECODER_REGISTRY


@TERRATORCH_DECODER_REGISTRY.register
class SatMAEHead(nn.Module):

    def __init__(self, embed_dim:int=None, out_channels:int=None, bias:bool=True,
            num_heads:int=None, mlp_ratio:float=None, patch_size:int=None, num_patches:int=1, 
                 depth:int=None, norm_layer=nn.LayerNorm, in_chans:int=None) -> None:

        super(SatMAEHead, self).__init__()

        if type(embed_dim)==tuple:
            self.embed_dim = embed_dim[1]
        else:
            self.embed_dim = embed_dim


        self.out_channels = out_channels
        self.out_channels = out_channels
        self.bias = bias
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.norm_layer = norm_layer
        self.in_chans = in_chans
        self.patch_size = patch_size
        self.num_patches = num_patches
        self.depth = depth 

        self.decoder_embed = nn.Linear(self.embed_dim, self.out_channels, bias=self.bias)

        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.out_channels))

        self.decoder_pos_embed = nn.Parameter(torch.zeros(1, self.num_patches + 1, self.out_channels), requires_grad=False)  # fixed sin-cos embedding

        self.decoder_blocks = nn.ModuleList([
            Block(self.out_channels, self.num_heads, self.mlp_ratio, qkv_bias=True, qk_norm=None, norm_layer=self.norm_layer)
            for i in range(self.depth)])

        self.decoder_norm = self.norm_layer(self.out_channels)
        self.decoder_pred = nn.Linear(self.out_channels, self.patch_size**2 * self.in_chans, bias=True) # decoder to patch

    def unpatchify(self, x, p, c):
        """
        x: (N, L, patch_size**2 *C)
        p: Patch embed patch size
        c: Num channels
        imgs: (N, C, H, W)
        """
        # c = self.in_c
        # p = self.patch_embed.patch_size[0]
        h = w = int(x.shape[1]**.5)
        assert h * w == x.shape[1]
        
        x = x.reshape(shape=(x.shape[0], h, w, p, p, c))
        x = torch.einsum('nhwpqc->nchpwq', x)
        imgs = x.reshape(shape=(x.shape[0], c, h * p, h * p))
        return imgs


    def forward(self, x, ids_restore) -> torch.Tensor:

        # embed tokens
        x = self.decoder_embed(x)

        # append mask tokens to sequence
        mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
        x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token

        # add pos embed
        x = x + self.decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)

        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)

        # remove cls token
        x = x[:, 1:, :]

        x = self.unpatchify(x, self.patch_size, self.in_chans)

        return x

class SatMAEHeadViT(nn.Module):

    def __init__(self, embed_dim:int=None, out_channels:int=None, bias:bool=True,
            num_heads:int=None, mlp_ratio:float=None, patch_size:int=None, num_patches:int=1, 
                 depth:int=None, norm_layer=nn.LayerNorm, in_chans:int=None) -> None:

        super(SatMAEHeadViT, self).__init__()

        if type(embed_dim)==tuple:
            self.embed_dim = embed_dim[1]
        else:
            self.embed_dim = embed_dim


        self.out_channels = out_channels
        self.bias = bias
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.norm_layer = norm_layer
        self.in_chans = in_chans
        self.patch_size = patch_size
        self.num_patches = num_patches
        self.depth = depth 

        self.decoder_embed = nn.Linear(self.embed_dim, self.out_channels, bias=self.bias)

        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.out_channels))

        self.decoder_pos_embed = nn.Parameter(torch.zeros(1, self.num_patches + 1, self.out_channels), requires_grad=False)  # fixed sin-cos embedding

        self.decoder_blocks = nn.ModuleList([
            Block(self.out_channels, self.num_heads, self.mlp_ratio, qkv_bias=True, qk_norm=None, norm_layer=self.norm_layer)
            for i in range(self.depth)])

        self.decoder_norm = self.norm_layer(self.out_channels)
        self.decoder_pred = nn.Linear(self.out_channels, self.patch_size**2 * self.in_chans, bias=True) # decoder to patch

    def unpatchify(self, x, p, c):
        """
        x: (N, L, patch_size**2 *C)
        p: Patch embed patch size
        c: Num channels
        imgs: (N, C, H, W)
        """
        # c = self.in_c
        # p = self.patch_embed.patch_size[0]
        h = w = int(x.shape[1]**.5)
        assert h * w == x.shape[1]
        
        x = x.reshape(shape=(x.shape[0], h, w, p, p, c))
        x = torch.einsum('nhwpqc->nchpwq', x)
        imgs = x.reshape(shape=(x.shape[0], c, h * p, h * p))
        return imgs


    def forward(self, x) -> torch.Tensor:

        x = torch.stack(x, dim=0)

        # embed tokens
        x = self.decoder_embed(x)

        # append mask tokens to sequence
        #mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        #x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        #x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
        #x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token

        # add pos embed
        #x = x + self.decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)

        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)

        # remove cls token
        x = x[:, 1:, :]

        x = self.unpatchify(x, self.patch_size, self.in_chans)

        return x




import torch
from torch import Tensor, nn
from torchgeo.trainers import BaseTask
from typing import Any, Mapping

from terratorch.models.model import Model#, get_factory
from terratorch.registry import MODEL_FACTORY_REGISTRY
from terratorch.tasks.loss_handler import LossHandler
from terratorch.tasks.optimizer_factory import optimizer_factory
from terratorch.tasks.regression_tasks import RootLossWrapper
from granitewxc.utils.config import ExperimentConfig
from torchmetrics import MetricCollection
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MetricCollection


class WxCDownscalingTask(BaseTask):
    def __init__(
        self,
        model_args: dict,
        model_factory: str,
        extra_kwargs: dict,
        model_config: ExperimentConfig,
        loss: str = "mse",
        lr: float = 0.001,
        optimizer: str | None = None,
        optimizer_hparams: dict | None = None,
        scheduler: str | None = None,
        scheduler_hparams: dict | None = None,
        freeze_backbone: bool = False,  # noqa: FBT001, FBT002
        freeze_decoder: bool = False,  # noqa: FBT001, FBT002
        plot_on_val: bool | int = 10,
    ) -> None:

        # Special cases for some parameters that could not be read in
        # their own fields.
        mask_unit_size = tuple(model_args.pop("mask_unit_size"))
        encoder_decoder_kernel_size_per_stage = extra_kwargs.pop("encoder_decoder_kernel_size_per_stage")
        output_vars = extra_kwargs.pop("output_vars")
        type_dataset = extra_kwargs.pop("type")
        input_levels = extra_kwargs.pop("input_levels")
        downscaling_patch_size = extra_kwargs.pop("downscaling_patch_size")
        n_input_timestamps = extra_kwargs.pop("n_input_timestamps")
        downscaling_embed_dim = extra_kwargs.pop("downscaling_embed_dim")
        encoder_decoder_conv_channels = extra_kwargs.pop("encoder_decoder_conv_channels")
        encoder_decoder_scale_per_stage = extra_kwargs.pop("encoder_decoder_scale_per_stage")
        encoder_decoder_upsampling_mode = extra_kwargs.pop("encoder_decoder_upsampling_mode")
        encoder_shift = extra_kwargs.pop("encoder_shift")
        drop_path = extra_kwargs.pop("drop_path")
        encoder_decoder_type = extra_kwargs.pop("encoder_decoder_type")
        input_size_lat = extra_kwargs.pop("input_size_lat")
        input_size_lon = extra_kwargs.pop("input_size_lon")
        freeze_backbone = extra_kwargs.pop("freeze_backbone")
        freeze_decoder = extra_kwargs.pop("freeze_decoder")
        data_path_surface = extra_kwargs.pop("data_path_surface") 
        data_path_vertical = extra_kwargs.pop("data_path_vertical") 
        climatology_path_surface = extra_kwargs.pop("climatology_path_surface") 
        climatology_path_vertical = extra_kwargs.pop("climatology_path_vertical") 
        residual_connection = model_args.pop("residual_connection")
        residual = extra_kwargs.pop("residual", True)

        # Special cases for required variables
        input_scalers_surface_path = extra_kwargs.pop("input_scalers_surface_path", None)
        if not input_scalers_surface_path:
            raise Exception(f"The parameter `input_scalers_surface_path` must be defined in `extra_kwargs`.")

        input_scalers_vertical_path = extra_kwargs.pop("input_scalers_vertical_path", None)
        if not input_scalers_vertical_path:
            raise Exception(f"The parameter `input_scalers_vertical_path` must be defined in `extra_kwargs`.")
        
        output_scalers_surface_path = extra_kwargs.pop("output_scalers_surface_path")
        output_scalers_vertical_path = extra_kwargs.pop("output_scalers_vertical_path")

        model_config.freeze_backbone = freeze_backbone
        model_config.freeze_decoder = freeze_decoder
        model_config.mask_unit_size = mask_unit_size  
        model_config.model.mask_unit_size = mask_unit_size
        model_config.model.encoder_decoder_kernel_size_per_stage = encoder_decoder_kernel_size_per_stage
        model_config.model.input_scalers_surface_path = input_scalers_surface_path
        model_config.model.input_scalers_vertical_path = input_scalers_vertical_path
        model_config.data.output_vars = output_vars
        model_config.data.type = type_dataset
        model_config.data.input_surface_vars = model_config.data.surface_vars
        model_config.data.input_vertical_vars = model_config.data.vertical_vars
        model_config.data.input_static_surface_vars = model_config.data.static_surface_vars
        model_config.data.input_levels = input_levels 
        model_config.model.downscaling_patch_size = downscaling_patch_size
        model_config.data.n_input_timestamps = n_input_timestamps
        model_config.model.downscaling_embed_dim = downscaling_embed_dim
        model_config.model.encoder_decoder_conv_channels = encoder_decoder_conv_channels
        model_config.model.encoder_decoder_scale_per_stage = encoder_decoder_scale_per_stage
        model_config.model.encoder_decoder_upsampling_mode = encoder_decoder_upsampling_mode
        model_config.model.encoder_shift = encoder_shift
        model_config.model.drop_path = drop_path 
        model_config.model.encoder_decoder_type = encoder_decoder_type
        model_config.data.input_size_lat = input_size_lat
        model_config.data.input_size_lon = input_size_lon
        model_config.data.data_path_surface = data_path_surface
        model_config.data.data_path_vertical = data_path_vertical
        model_config.data.climatology_path_surface = climatology_path_surface
        model_config.data.climatology_path_vertical = climatology_path_vertical
        model_config.model.output_scalers_surface_path = output_scalers_surface_path
        model_config.model.output_scalers_vertical_path = output_scalers_vertical_path
        model_config.model.residual_connection = residual_connection
        model_config.model.residual = residual

        self.model_factory = MODEL_FACTORY_REGISTRY.build(model_factory)
        self.model_config = model_config
        # TODO Unify it with self.hparams
        self.extended_hparams = self.model_config.to_dict()
        super().__init__()

        self.train_loss_handler = LossHandler(self.train_metrics.prefix)
        self.test_loss_handler = LossHandler(self.test_metrics.prefix)
        self.val_loss_handler = LossHandler(self.val_metrics.prefix)
        self.monitor = f"{self.val_metrics.prefix}loss"
        self.plot_on_val = int(plot_on_val)

    def configure_models(self) -> None:
        self.model: Model = self.model_factory.build_model(
            "regression", aux_decoders=None, model_config=self.model_config, **self.hparams["model_args"]
        )
        if self.extended_hparams["freeze_backbone"]:
            self.model.freeze_encoder()
        if self.extended_hparams["freeze_decoder"]:
            self.model.freeze_decoder()

    def configure_optimizers(
        self,
    ) -> "lightning.pytorch.utilities.types.OptimizerLRSchedulerConfig":
        optimizer = self.hparams["optimizer"]
        if optimizer is None:
            optimizer = "Adam"
        return optimizer_factory(
            self.hparams["optimizer"],
            self.hparams["lr"],
            self.parameters(),
            self.hparams["optimizer_hparams"],
            self.hparams["scheduler"],
            self.monitor,
            self.hparams["scheduler_hparams"],
        )

    def configure_losses(self) -> None:
        """Initialize the loss criterion.

        Raises:
            ValueError: If *loss* is invalid.
        """
        #TODO 'reduction' should be chosen using the config and
        # a similar class as IgnoreIndex should be defined for this class

        loss: str = self.hparams["loss"].lower()
        if loss == "mse":
            self.criterion: nn.Module = nn.MSELoss(reduction="mean")
        elif loss == "mae":
            self.criterion = nn.L1Loss(reduction="mean")
        elif loss == "rmse":
            # IMPORTANT! Root is done only after ignore index! Otherwise the mean taken is incorrect
            self.criterion = RootLossWrapper(nn.MSELoss(reduction="none"), reduction="none")
        elif loss == "huber":
            self.criterion = nn.HuberLoss(reduction="none")
        else:
            exception_message = f"Loss type '{loss}' is not valid. Currently, supports 'mse', 'rmse' or 'mae' loss."
            raise ValueError(exception_message)

    def configure_metrics(self) -> None:

        def instantiate_metrics():
            return {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
            }

        self.train_metrics = MetricCollection(instantiate_metrics(), prefix="train/")
        self.val_metrics = MetricCollection(instantiate_metrics(), prefix="val/")
        self.test_metrics = MetricCollection(instantiate_metrics(), prefix="test/")

    def training_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Tensor:

        x = batch["image"]
        mask = batch["mask"]
        model_output: ModelOutput = self(x)
        y = mask
        loss = self.train_loss_handler.compute_loss(model_output, y, self.criterion, None)
        self.train_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = model_output.output
        self.train_metrics(y_hat, y)
        self.log_dict(self.train_metrics, on_epoch=True)

        return loss["loss"]

    def validation_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the train loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        y = batch["mask"]
        model_output: ModelOutput = self(x)

        loss = self.val_loss_handler.compute_loss(model_output, y, self.criterion, None)
        self.val_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = model_output.output
        self.val_metrics(y_hat, y)
        self.log_dict(self.val_metrics, on_epoch=True)

        return loss["loss"]

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        raise NotImplementedError("This function is not yet implemented.")

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        x = batch["image"]
        file_names = batch["filename"]

        y_hat: Tensor = self(x).output
        return y_hat, file_names



# Copyright contributors to the Terratorch project

import importlib

import pytest
import torch

from terratorch.models import EncoderDecoderFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS
from terratorch.models.model import AuxiliaryHead
from terratorch.models.backbones import torchgeo_resnet as torchgeo_resnet

NUM_CHANNELS = 6
NUM_CLASSES = 10
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)
EXPECTED_REGRESSION_OUTPUT_SHAPE = (1, 224, 224)
EXPECTED_CLASSIFICATION_OUTPUT_SHAPE = (1, NUM_CLASSES)

PIXELWISE_TASK_EXPECTED_OUTPUT = [
    ("regression", EXPECTED_REGRESSION_OUTPUT_SHAPE),
    ("segmentation", EXPECTED_SEGMENTATION_OUTPUT_SHAPE),
]

VIT_UPERNET_NECK = [
    {"name": "SelectIndices", "indices": [0, 1, 2, 3]},
    {"name": "ReshapeTokensToImage"},
    {"name": "LearnedInterpolateToPyramidal"},
]

PRETRAINED_BANDS = ["RED", "GREEN", "BLUE", "NIR_NARROW", "SWIR_1", "SWIR_2"]

@pytest.fixture(scope="session")
def model_factory() -> EncoderDecoderFactory:
    return EncoderDecoderFactory()


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))

def torchgeo_resnet_backbones():
    return [i for i in dir(torchgeo_resnet) if "_resnet" in i and i != "load_resnet_weights"]

backbones = torchgeo_resnet_backbones()
pretrained = [False]
@pytest.mark.parametrize("backbone", backbones)
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_classification_model_resnet(backbone, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_model_bands=PRETRAINED_BANDS,
        backbone_pretrained=backbone_pretrained,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

backbones = ["ssl4eos12_resnet50_sentinel2_all_decur"]
pretrained = [True]
@pytest.mark.parametrize("backbone", backbones)
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_classification_model_resnet_pretrained(backbone, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_model_bands=PRETRAINED_BANDS,
        backbone_pretrained=backbone_pretrained,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

backbones = ["dofa_large_patch16_224"]
@pytest.mark.parametrize("backbone", backbones)
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_classification_model_dofa(backbone, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_model_bands=PRETRAINED_BANDS,
        backbone_pretrained=backbone_pretrained,
        num_classes=NUM_CLASSES,
        necks = [{"name": "PermuteDims", "new_order": [0, 2, 1]}]
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

backbones = ["satlas_swin_b_sentinel2_si_ms"]
@pytest.mark.parametrize("backbone", backbones)
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_classification_model_swin(backbone, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_model_bands=PRETRAINED_BANDS,
        backbone_pretrained=backbone_pretrained,
        num_classes=NUM_CLASSES,
        necks = [{"name": "PermuteDims", "new_order": [0, 3, 1, 2]}]
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["ssl4eos12_resnet50_sentinel2_all_decur"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "IdentityDecoder", "smp_Unet"])
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_pixelwise_model_resnet(backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_model_bands": PRETRAINED_BANDS,
        "backbone_pretrained": backbone_pretrained,
        "backbone_out_indices": [0, 1, 2, 3, 4], 
        
    }
        
    if decoder == "smp_Unet":
        model_args["decoder_decoder_channels"] = [512, 256, 128, 64]
    
    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected



@pytest.mark.parametrize("backbone", ["dofa_large_patch16_224"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_pixelwise_model_dofa(backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_model_bands": PRETRAINED_BANDS,
        "backbone_pretrained": backbone_pretrained,
        "backbone_out_indices": [5, 11, 17, 23]
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
        
    model_args["necks"] = [{"name": "ReshapeTokensToImage"}]

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected


@pytest.mark.parametrize("backbone", ["satlas_swin_b_sentinel2_si_ms"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["UperNetDecoder", "IdentityDecoder"])
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_pixelwise_model_swin(backbone, task, expected, decoder, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "backbone_model_bands": PRETRAINED_BANDS,
        "backbone_pretrained": backbone_pretrained,
        "backbone_out_indices": [1, 3, 5, 7]
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
        
    model_args["necks"] = [{"name": "PermuteDims", "new_order": [0, 3, 1, 2]}]

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected


